@@ -7,8 +7,9 @@ from hypothesis import HealthCheck, given, settings
 from hypothesis import strategies as st
 from pydantic import ValidationError
 from pytest_httpserver import HTTPServer
+from pytest_mock import MockerFixture
 
-from harborapi.client import HarborAsyncClient, construct_model
+from harborapi.client import HarborAsyncClient
 from harborapi.exceptions import (
     BadRequest,
     Forbidden,
@@ -174,6 +175,53 @@ async def test_get_pagination_no_follow(
     assert users[1]["username"] == "user2"
 
 
+@pytest.mark.asyncio
+async def test_get_pagination_limit(
+    async_client: HarborAsyncClient, httpserver: HTTPServer
+):
+    """Test pagination by mocking paginated /users results."""
+    httpserver.expect_oneshot_request("/api/v2.0/users").respond_with_json(
+        [{"username": "user1"}, {"username": "user2"}],
+        headers={"link": '</api/v2.0/users?page=2>; rel="next"'},
+    )
+    httpserver.expect_oneshot_request(
+        "/api/v2.0/users", query_string=f"page=2"
+    ).respond_with_json(
+        [{"username": "user3"}],
+        headers={"link": '</api/v2.0/users?page=1>; rel="prev"'},
+    )
+    async_client.url = httpserver.url_for("/api/v2.0")
+    users = await async_client.get("/users", limit=2)  # type: ignore
+    assert isinstance(users, list)
+    assert len(users) == 2
+    assert users[0]["username"] == "user1"
+    assert users[1]["username"] == "user2"
+
+
+@pytest.mark.asyncio
+async def test_get_pagination_no_limit(
+    async_client: HarborAsyncClient, httpserver: HTTPServer
+):
+    """Test pagination by mocking paginated /users results."""
+    httpserver.expect_oneshot_request("/api/v2.0/users").respond_with_json(
+        [{"username": "user1"}, {"username": "user2"}],
+        headers={"link": '</api/v2.0/users?page=2>; rel="next"'},
+    )
+    httpserver.expect_oneshot_request(
+        "/api/v2.0/users", query_string=f"page=2"
+    ).respond_with_json(
+        [{"username": "user3"}],
+        headers={"link": '</api/v2.0/users?page=1>; rel="prev"'},
+    )
+    async_client.url = httpserver.url_for("/api/v2.0")
+    users = await async_client.get("/users", limit=None)  # type: ignore
+    assert isinstance(users, list)
+    assert len(users) == 3
+    assert users[0]["username"] == "user1"
+    assert users[1]["username"] == "user2"
+    assert users[2]["username"] == "user3"
+
+
 @pytest.mark.asyncio
 async def test_get_pagination_invalid_mock(
     async_client: HarborAsyncClient,
@@ -225,17 +273,135 @@ async def test_get_retry_mock(async_client: HarborAsyncClient, httpserver: HTTPS
 
 
 @pytest.mark.asyncio
-async def test_construct_model():
+async def test_construct_model(async_client: HarborAsyncClient, mocker: MockerFixture):
+    c = async_client
+    construct_spy = mocker.spy(UserResp, "construct")
+    parse_spy = mocker.spy(UserResp, "parse_obj")
+
     # TODO: test create_model with all models
-    m = construct_model(UserResp, {"username": "user1"})
+    m = c.construct_model(UserResp, {"username": "user1"})
     assert isinstance(m, UserResp)
     assert m.username == "user1"
+    assert parse_spy.call_count == 1
+    assert construct_spy.call_count == 0
 
     # Invalid value for "username"
     with pytest.raises(ValidationError) as e:
-        construct_model(UserResp, {"username": {}})
+        c.construct_model(UserResp, {"username": {}})
     assert e.value.errors()[0]["loc"] == ("username",)
     assert len(e.value.errors()) == 1
+    assert parse_spy.call_count == 2
+    assert construct_spy.call_count == 0
+
+
+@pytest.mark.asyncio
+async def test_construct_model_no_validation(
+    async_client: HarborAsyncClient,
+    mocker: MockerFixture,
+):
+    c = async_client
+    c.validate = False
+    construct_spy = mocker.spy(UserResp, "construct")
+    parse_spy = mocker.spy(UserResp, "parse_obj")
+
+    # Extra field "foo" is added to the model
+    m = c.construct_model(UserResp, {"username": "user1", "foo": "bar"})
+    assert isinstance(m, UserResp)
+    assert m.username == "user1"
+    assert m.foo == "bar"
+    assert construct_spy.call_count == 1
+    assert parse_spy.call_count == 0
+
+    # Invalid value for "username" if validation was enabled
+    m = c.construct_model(UserResp, {"username": {}})
+    assert m.username == {}
+    assert construct_spy.call_count == 2
+    assert parse_spy.call_count == 0
+
+
+@pytest.mark.asyncio
+async def test_construct_model_raw(
+    async_client: HarborAsyncClient,
+    mocker: MockerFixture,
+):
+    c = async_client
+    c.raw = True
+    construct_spy = mocker.spy(UserResp, "construct")
+    parse_spy = mocker.spy(UserResp, "parse_obj")
+
+    # Extra field "foo" is added to the model
+    expect_resp = {"username": "user1", "foo": "bar"}
+    m = c.construct_model(UserResp, expect_resp)
+    assert m == expect_resp
+    assert m["username"] == "user1"
+    assert m["foo"] == "bar"
+    assert construct_spy.call_count == 0
+    assert parse_spy.call_count == 0
+
+
+@pytest.mark.asyncio
+async def test_construct_model_raw_is_list(
+    async_client: HarborAsyncClient,
+    mocker: MockerFixture,
+):
+    c = async_client
+    c.raw = True
+    construct_spy = mocker.spy(UserResp, "construct")
+    parse_spy = mocker.spy(UserResp, "parse_obj")
+
+    expect_resp = [{"username": "user1", "foo": "bar"}, {"username": "user2"}]
+    m = c.construct_model(UserResp, expect_resp, is_list=True)
+    assert m == expect_resp
+    assert m[0]["username"] == "user1"
+    assert m[0]["foo"] == "bar"
+    assert m[1]["username"] == "user2"
+    assert construct_spy.call_count == 0
+    assert parse_spy.call_count == 0
+
+
+@pytest.mark.asyncio
+async def test_construct_model_raw_list_without_is_list(
+    async_client: HarborAsyncClient,
+    mocker: MockerFixture,
+):
+    """Receives list value, but is_list is set to False.
+    Since we are in raw mode, this is not an error.
+    """
+    c = async_client
+    c.raw = True
+    construct_spy = mocker.spy(UserResp, "construct")
+    parse_spy = mocker.spy(UserResp, "parse_obj")
+
+    expect_resp = [{"username": "user1", "foo": "bar"}, {"username": "user2"}]
+    m = c.construct_model(UserResp, expect_resp, is_list=False)
+    assert m == expect_resp
+    assert m[0]["username"] == "user1"
+    assert m[0]["foo"] == "bar"
+    assert m[1]["username"] == "user2"
+    assert construct_spy.call_count == 0
+    assert parse_spy.call_count == 0
+
+
+@pytest.mark.asyncio
+async def test_construct_model_raw_is_list_without_list(
+    async_client: HarborAsyncClient,
+    mocker: MockerFixture,
+):
+    """is_list is set to True, but receives non-list value.
+    Since we are in raw mode, this is not an error.
+    """
+    c = async_client
+    c.raw = True
+    construct_spy = mocker.spy(UserResp, "construct")
+    parse_spy = mocker.spy(UserResp, "parse_obj")
+
+    expect_resp = {"username": "user1", "foo": "bar"}
+    m = c.construct_model(UserResp, expect_resp, is_list=True)
+    assert m == expect_resp
+    assert m["username"] == "user1"
+    assert m["foo"] == "bar"
+    assert construct_spy.call_count == 0
+    assert parse_spy.call_count == 0
 
 
 @pytest.mark.asyncio
